
#from .vcf_ref_to_seq import *
#from .vcf_from_regions import *
#from .vcf_reader_func import *
#from .gene_region import *
#from .logging_module import *
#from .four_gamete_pysam import *
#from .vcf_to_ima import *
#from .find_intergenic_bed import *
#from .get_nonmissing_chunks import *
#from .vcf_phase import *
#from .vcf_calc import *
#from .vcf_filter import *
#from .vcf_split import *
#from .vcftools import *
#from .bcftools import *

# Basic Information
__title__     = "py-popgen"
__version__   = '0.1.2'
__summary__   = "Software platform for facilitating population genomic analyses"
__url__       = "https://ppp.csusm.edu/"
__license__   = "MIT"
__copyright__ = "Copyright 2019, Arun Sethuraman and Jody Hey"

# Author Information
__author__    = "Andrew Webb, Jared Knoblauch, Nitesh Sabankar, " \
                "Apeksha Sukesh Kallur, Jody Hey and Arun Sethuraman"               
__email__     = "asethuraman@csusm.edu"




